from firebolt.async_db.connection import Connection

from firebolt_ingest.aws_settings import (
    AWSSettings,
    generate_aws_credentials_string,
)
from firebolt_ingest.table_model import Table
from firebolt_ingest.table_utils import (
    drop_table,
    get_table_columns,
    get_table_schema,
)

EXTERNAL_TABLE_PREFIX = "ex_"


class TableService:
    def __init__(self, connection: Connection):
        """
        Table service class used for creation of external/internal tables and
        performing ingestion from external into internal table

        Args:
            connection: a connection to some database/engine
        """
        self.connection = connection

    def create_external_table(self, table: Table, aws_settings: AWSSettings) -> None:
        """
        Constructs a query for creating an external table and executes it.

        Args:
            table: table definition
            aws_settings: aws settings
        """
        # Prepare aws credentials
        if aws_settings.aws_credentials:
            cred_stmt, cred_params = generate_aws_credentials_string(
                aws_settings.aws_credentials
            )
            cred_stmt = f"CREDENTIALS = {cred_stmt}"
        else:
            cred_stmt, cred_params = "", []

        # Prepare columns
        columns_stmt, columns_params = table.generate_external_columns_string()

        # Generate query
        query = (
            f"CREATE EXTERNAL TABLE {EXTERNAL_TABLE_PREFIX}{table.table_name}\n"
            f"({columns_stmt})\n"
            f"{cred_stmt}\n"
            f"URL = ?\n"
            f"OBJECT_PATTERN = {', '.join(['?'] * len(table.object_pattern))}\n"
            f"TYPE = ({table.generate_file_type()})\n"
        )
        if table.compression:
            query += f"COMPRESSION = {table.compression}\n"

        params = (
            cred_params + columns_params + [aws_settings.s3_url] + table.object_pattern
        )

        # Execute parametrized query
        self.connection.cursor().execute(query, params)

    def create_internal_table(self, table: Table, add_file_metadata=True) -> None:
        """
        Constructs a query for creating an internal table and executes it

        Args:
            table: table definition
        """

        columns_stmt, columns_params = table.generate_internal_columns_string(
            add_file_metadata
        )
        query = (
            f"CREATE FACT TABLE {table.table_name}\n"
            f"({columns_stmt})\n"
            f"PRIMARY INDEX {table.generate_primary_index_string()}\n"
        )

        if table.partitions:
            query += f"PARTITION BY {table.generate_partitions_string(add_file_metadata=add_file_metadata)}\n"  # noqa: E501

        self.connection.cursor().execute(query, columns_params)

    def insert_full_overwrite(
        self,
        internal_table_name: str,
        external_table_name: str,
        firebolt_dont_wait_for_upload_to_s3: bool = False,
    ) -> None:
        """
        Perform a full overwrite from an external table into an internal table.
        All existing data in the internal table will be deleted, and data from the
        external table will be loaded.

        This function is appropriate for unpartitioned tables, or for
        partitioned tables you wish to fully overwrite.

        Args:
            internal_table_name: (destination) The internal table which
                                 will be overwritten.
            external_table_name: (source) The external table from which to load.

            firebolt_dont_wait_for_upload_to_s3: (Optional) if set, the insert will not
                wait until the changes will be written to s3.

        """
        cursor = self.connection.cursor()

        # get table schema
        internal_table_schema = get_table_schema(cursor, internal_table_name)

        # drop the table
        drop_table(cursor, internal_table_name)

        # recreate the table
        cursor.execute(query=internal_table_schema)

        # insert the data from external to internal
        column_names = get_table_columns(cursor, internal_table_name)
        insert_query = (
            f"INSERT INTO {internal_table_name}\n"
            f"SELECT {', '.join(column_names)}\n"
            f"FROM {external_table_name}\n"
        )

        if firebolt_dont_wait_for_upload_to_s3:
            cursor.execute(query="set firebolt_dont_wait_for_upload_to_s3=1")

        cursor.execute(query=insert_query)



from functools import wraps
from typing import Any, List, Optional, Sequence

from firebolt.async_db import Cursor
from firebolt.common.exception import FireboltError


def table_must_exist(func):
    @wraps(func)
    def with_table_existence_check(cursor: Cursor, table_name: str, **kwargs):
        if not does_table_exist(cursor=cursor, table_name=table_name):
            raise FireboltError(
                f"Table {table_name} does not exist " f"when calling {func.__name__}"
            )
        return func(cursor=cursor, table_name=table_name, **kwargs)

    return with_table_existence_check


def get_table_schema(cursor: Cursor, table_name: str) -> str:
    """
    Return the create command of the existing table.

    Args:
        cursor: Firebolt database cursor
        table_name: Name of the table

    Returns:
        CREATE TABLE ... command
    """
    cursor.execute("SHOW TABLES")
    data = cursor.fetchall()

    def find_column_index(columns: Sequence, column_name: str) -> int:
        result = [
            idx for idx, column in enumerate(columns) if column.name == column_name
        ]

        if len(result) == 0:
            raise FireboltError(f"Cannot find expected column: {column_name}")
        elif len(result) > 1:
            raise FireboltError(f"Too many columns: {column_name} found")

        return result[0]

    table_name_index = find_column_index(cursor.description, "table_name")
    schema_index = find_column_index(cursor.description, "schema")

    internal_table_schema = [
        row[schema_index]
        for row in data  # type: ignore
        if row[table_name_index] == table_name
    ]

    if len(internal_table_schema) == 0:
        raise FireboltError("internal table doesn't exist")

    return internal_table_schema[0]


def drop_table(cursor: Cursor, table_name: str) -> None:
    """
    Drop a table.

    Args:
        cursor: Firebolt database cursor
        table_name: Name of the table to drop.

    Raises an exception if the table did not drop.
    """
    drop_query = f"DROP TABLE IF EXISTS {table_name} CASCADE"

    # drop the table
    cursor.execute(query=drop_query)

    # verify that the drop succeeded
    if does_table_exist(cursor, table_name):
        raise FireboltError(f"Table {table_name} did not drop successfully.")


@table_must_exist
def get_table_columns(cursor: Cursor, table_name: str) -> List[str]:
    """
    Get the column names of an existing table on Firebolt.

    Args:
        cursor: Firebolt database cursor
        table_name: Name of the table
    """
    cursor.execute(f"SELECT * FROM {table_name} LIMIT 0")
    return [column.name for column in cursor.description]


@table_must_exist
def get_table_partition_columns(cursor: Cursor, table_name: str) -> List[str]:
    """
    Get the names of partition columns of an existing table on Firebolt.

    Args:
        cursor: Firebolt database cursor
        table_name: Name of the table.
    """
    sql = """
    SELECT column_name
    FROM information_schema.columns
    WHERE
        table_schema = ? AND
        table_name = ? AND
        is_in_partition_expr = 'YES'
    """
    cursor.execute(query=sql, parameters=(cursor.connection.database, table_name))
    return cursor.fetchall()  # type: ignore


@table_must_exist
def get_partition_keys(
    cursor: Cursor, table_name: str, where_sql: Optional[str] = None
) -> Sequence[Any]:
    """
    Get the partition keys of an existing table on Firebolt.

    Args:
        cursor: Firebolt database cursor
        table_name: Name of the table.
        where_sql: (Optional) A where clause, for filtering data.
            Do not include the "WHERE" keyword.
            If no clause is provided (default), all partition keys are returned.
    """
    # FUTURE: replace this query with `show partitions` when
    # https://packboard.atlassian.net/browse/FIR-2370 is completed
    part_expr = ",".join(
        get_table_partition_columns(cursor=cursor, table_name=table_name)
    )
    sql = f"""
    SELECT DISTINCT {part_expr}
    FROM {table_name}
    """
    if where_sql is not None:
        sql += f"WHERE {where_sql}"

    cursor.execute(sql)
    return cursor.fetchall()  # type: ignore


def does_table_exist(cursor: Cursor, table_name: str) -> bool:
    """
    Check whether table with table_name exists,
    and return True if it exists, False otherwise.
    """
    find_query = f"SELECT * FROM information_schema.tables WHERE table_name = ?"

    return cursor.execute(find_query, [table_name]) != 0



from enum import Enum
from typing import Any, List, Optional, Tuple

import yaml
from pydantic import BaseModel, Field, ValidationError, conlist, root_validator
from pydantic.main import ModelMetaclass
from yaml import Loader


class YamlModelMixin(metaclass=ModelMetaclass):
    """
    Provides a parse_yaml method to a Pydantic model class.
    """

    @classmethod
    def parse_yaml(cls, yaml_obj: Any):
        obj = yaml.load(yaml_obj, Loader=Loader)
        return cls.parse_obj(obj)  # type: ignore


# see: https://docs.firebolt.io/general-reference/data-types.html
ATOMIC_TYPES = {
    "INT",
    "INTEGER",
    "BIGINT",
    "LONG",
    "FLOAT",
    "DOUBLE",
    "DOUBLE PRECISION",
    "TEXT",
    "VARCHAR",
    "STRING",
    "DATE",
    "DATETIME",
    "TIMESTAMP",
    "BOOLEAN",
}

DATE_TIME_TYPES = {"DATE", "TIMESTAMP", "DATETIME"}


def match_array(s: str) -> bool:
    """
    Check whether a string is a valid array and return true
    """
    if s.startswith("ARRAY(") and s.endswith(")"):
        return match_array(s[6:-1])
    if s in ATOMIC_TYPES:
        return True
    return False


class DatetimePart(str, Enum):
    DAY = "DAY"
    DOW = "DOW"
    MONTH = "MONTH"
    WEEK = "WEEK"
    WEEKISO = "WEEKISO"
    QUARTER = "QUARTER"
    YEAR = "YEAR"
    HOUR = "HOUR"
    MINUTE = "MINUTE"
    SECOND = "SECOND"
    EPOCH = "EPOCH"


class Column(BaseModel):
    name: str = Field(min_length=1, max_length=255, regex=r"^[0-9a-zA-Z_]+$")
    type: str = Field(min_length=1, max_length=255)
    extract_partition: Optional[str] = Field(min_length=1, max_length=255)
    nullable: Optional[bool] = None
    unique: Optional[bool] = None

    @root_validator
    def type_validator(cls, values: dict) -> dict:
        if values["type"] in ATOMIC_TYPES or match_array(values["type"]):
            return values

        raise ValueError("unknown column type")


FILE_METADATA_COLUMNS: List[Column] = [
    Column(name="source_file_name", type="STRING"),
    Column(name="source_file_timestamp", type="DATETIME"),
]


class Partition(BaseModel):
    """
    A partition column or expression.
    If a DatetimePart is provided, it will be an expression.
    Currently, the only supported expressions are date/time EXTRACT(...)

    see: https://docs.firebolt.io/working-with-partitions.html
    see: https://docs.firebolt.io/sql-reference/functions-reference/date-and-time-functions.html#extract  # noqa: E501
    """

    column_name: str
    datetime_part: Optional[DatetimePart]

    def as_sql_string(self) -> str:
        if self.datetime_part is not None:
            return f"EXTRACT({self.datetime_part} FROM {self.column_name})"
        return self.column_name


class Table(BaseModel, YamlModelMixin):
    table_name: str = Field(min_length=1, max_length=255, regex=r"^[0-9a-zA-Z_]+$")
    columns: conlist(Column, min_items=1)  # type: ignore
    primary_index: conlist(str, min_items=1)  # type: ignore
    partitions: List[Partition] = []
    file_type: str
    object_pattern: conlist(str, min_items=1)  # type: ignore
    compression: Optional[str] = None
    csv_skip_header_row: Optional[bool] = None
    json_parse_as_text: Optional[bool] = None

    @root_validator
    def compression_validator(cls, values: dict) -> dict:
        """
        Check whether compression has one of allowed values: {"GZIP"}
        """
        if values.get("compression") is None:
            return values

        values["compression"] = values["compression"].upper()

        if values.get("compression") not in {"GZIP"}:
            raise ValidationError(["unknown compression"], type(cls))

        return values

    @root_validator
    def file_type_validator(cls, values: dict) -> dict:
        """
        Check whether file_type has one of allowed values:
            {"CSV", "JSON", "ORC", "PARQUET", "TCV"}
        """
        values["file_type"] = values["file_type"].upper()

        if values.get("file_type") not in {"CSV", "JSON", "ORC", "PARQUET", "TCV"}:
            raise ValidationError(["unknown file type"], type(cls))

        return values

    @root_validator
    def file_type_additional_fields(cls, values: dict) -> dict:
        """
        validate that if csv_skip_header_row is set, then the file type is CSV
        and if json_parse_as_text is set, the file type is JSON
        """
        if values.get("csv_skip_header_row") and values.get("file_type") != "CSV":
            raise ValidationError(
                ["csv_skip_header_row is only relevant for CSV file_type"], type(cls)
            )

        if values.get("json_parse_as_text") and values.get("file_type") != "JSON":
            raise ValidationError(
                ["json_parse_as_text is only relevant for JSON file_type"], type(cls)
            )

        return values

    @root_validator
    def primary_index_columns(cls, values: dict) -> dict:
        """
        Ensure the primary index column names exist in the list of columns.
        """
        column_names = {c.name for c in values["columns"]}
        for index_column in values.get("primary_index", []):
            if index_column not in column_names:
                raise ValidationError(
                    [
                        f"Could not find primary index {index_column}"
                        f" in the list of table columns."
                    ],
                    type(cls),
                )
        return values

    @root_validator
    def partition_columns(cls, values: dict) -> dict:
        """
        Ensure the partition column_name exists in the list of columns.
        Ensure partition columns that use EXTRACT refer to date/time columns.
        """
        column_name_to_type = {c.name: c.type for c in values["columns"]}
        for partition in values.get("partitions", []):

            if partition.column_name not in column_name_to_type.keys():
                raise ValidationError(
                    [
                        f"Could not find partition column name {partition.column_name} "
                        f"in the list of table columns"
                    ],
                    type(cls),
                )

            if partition.datetime_part is not None:
                partition_column_type = column_name_to_type.get(partition.column_name)
                if partition_column_type not in DATE_TIME_TYPES:
                    raise ValidationError(
                        [
                            f"Partition column {partition.column_name} must be a "
                            f"compatible datetime type, not a {partition_column_type}"
                        ],
                        type(cls),
                    )
        return values

    def generate_file_type(self) -> str:
        """
        Returns: a string with file_type and relevant argument
        """
        additional_params = ""
        if self.file_type == "CSV" and self.csv_skip_header_row is not None:
            additional_params = (
                f" SKIP_HEADER_ROWS = {'1' if self.csv_skip_header_row else '0'}"
            )
        elif self.file_type == "JSON" and self.json_parse_as_text is not None:
            additional_params = (
                f" PARSE_AS_TEXT = '{'TRUE' if self.json_parse_as_text else 'FALSE'}'"
            )

        return f"{self.file_type}{additional_params}"

    def generate_internal_columns_string(
        self, add_file_metadata: bool
    ) -> Tuple[str, List]:
        """

        Generate a prepared sql string from list of columns to
        be used in the creation of internal tables.

        Args:
            add_file_metadata: If true, add the source_file_name and
            source_file_timestamp to the list of columns.

        Returns:
            a tuple with partial sql prepared statement and
            list of prepared statement arguments
        """
        additional_partitions = FILE_METADATA_COLUMNS if add_file_metadata else []

        columns_str = []
        for column in self.columns + additional_partitions:
            column_str = f"{column.name} {column.type}"

            column_str += " UNIQUE" if column.unique else ""

            if column.nullable is not None:
                column_str += " NULL" if column.nullable else " NOT NULL"

            columns_str.append(column_str)

        return ", ".join(columns_str), []

    def generate_external_columns_string(self) -> Tuple[str, List]:
        """
        Generate a prepared sql string from list of columns to
        be used in the creation of external table.

        Returns:
            a tuple with partial sql prepared statement and
            list of prepared statement arguments
        """

        column_strings = []
        for column in self.columns:
            column_strings.append(
                f"{column.name} {column.type}"
                f"{' PARTITION(?)' if column.extract_partition else ''}"
            )

        return ", ".join(column_strings), [
            column.extract_partition
            for column in self.columns
            if column.extract_partition
        ]

    def generate_primary_index_string(self) -> str:
        """
        Generate a prepared sql string from list of primary index columns to
        be used in the creation of internal tables with a primary index.
        """
        return ", ".join([index for index in self.primary_index])

    def generate_partitions_string(self, add_file_metadata: bool) -> str:
        """
        Generate a prepared sql string from list of partition columns to
        be used in the creation of internal partitioned tables.

        Args:
            add_file_metadata: If true, add the source_file_name and
            source_file_timestamp columns as partition columns.
        """
        additional_partitions = (
            [Partition(column_name=c.name) for c in FILE_METADATA_COLUMNS]
            if add_file_metadata
            else []
        )
        return ",".join(
            [p.as_sql_string() for p in self.partitions + additional_partitions]
        )
